/* Filename: SROCK.c */
/* 27-Feb-2011 */

#include <stdio.h>
#include <math.h>

#define MaxStageNum 104 /* Maximum stage number */

#define SQ3 1.7320508075688773 /* sqrt(3) */

#define MaxCoreNum 8 /* 8, 1 *//* Maximum number of multi-core. */
/*
  In this program code, OpenMP is available. If you want to use it,
  remove a comment out in the line "#pragma omp parallel for".
 */

int GetSRKVal(int ss, double Mu[], double Ka[], double *sig, double *coe,
	      double *al1sM1sM2, double *al1ssM2, double *al1ssM1,
	      double *al2sM2sM3, double *al2sM1sM3, double *al2ssM3,
	      double *al2ssM2, double *al2ssM1);

extern int OMP_wo2_srock_for_NonCommSDEs_WinMulti(int ydim,
						  unsigned long traj,
						  double *yvec,
						  double step,
						  char *ran3p, char *ran2p,
						  void (*ffunc)(),
						  void (*gfunc_gene)(),
						  int wdim, int ss,
						  double work[], double *ynew)
/* This function performs SROCK2 scheme based on ROCK2.
   It gives all trajectries for one step concerning SDEs with
   a multi-dimentional Winer process. If an error occurs, it will
   return 1, otherwise 0.

   Input arguments
   ----------------
   ydim: dimension of SDEs,
   traj: number of trajectries,
   yvec: pointer of the head of all initial values, which are in order
         like yvec[0], yvec[1], ..., yvec[ydim-1] for the 1st trajectry,
	 yvec[ydim], yvec[ydim+1], ..., yvec[2*ydim-1] for the 2nd trajectry.
   step: step length,
   ran3p: pointer of the head of 3*traj three-point distributed RVs
          with P(-1)=P(1)=1/6 and P(0)=2/3,
   ran2p: pointer of the head of 2*traj two-point distributed RVs
          with P(-1)=P(1)=1/2,
   ffunc: drift coefficient,
   gfunc_gene: diffusion coefficients,
   wdim: dimension of Wiener process.
   ss: stage numer of SRK scheme.

   Workspace arguments
   -------------------
   work: workspace of length MaxCoreNum*(8+7*wdim+wdim*(wdim-1))*ydim.

   Output arguments
   ----------------
   ynew: pointer of the head of all solutions for one step, which are
         in a similar order to yvec.
*/
{
  static int static_flag=0;
  static unsigned long traj_mini;
  int errflag, sM2, sM3, wdimM1, wdimM2, ii_par;
  /* See Section 4 in [Burrage_and_Komori:2010]. */
  static double   b1sM3=1.0/8.0, b1sM2=3.0/8.0, b1sM1=3.0/8.0, b1s=1.0/8.0,
    al3sM2sM3=2.0/3.0, al3sM1sM3=1.0/12.0, al3sM1sM2=1.0/4.0,
    al3ssM3=-5.0/4.0, al3ssM2=1.0/4.0, al3ssM1=2.0,
    al4sM1sM3=1.0/4.0, al4sM1sM2=3.0/4.0, al4ssM3=1.0/4.0, al4ssM2=3.0/4.0;
  double Mu[MaxStageNum-2], Ka[MaxStageNum-3], sigma, coeff,
    al1sM1sM2, al1ssM2, al1ssM1,
    al2sM2sM3, al2sM1sM3, al2sM1sM2, al2ssM3, al2ssM2, al2ssM1,
    al0sM1sM2, al6ssM2, al6sM1sM2, b2sM1, b2s, sqstep, sq3step;
  
  if((ydim <=0) || (step <= 0))
    {printf("negative argument err\n");return 1;}/* Check for arguments */
  sqstep = sqrt(step);
  sq3step=SQ3*sqstep;
  
  wdimM1=wdim-1;
  wdimM2=wdim-2;

  errflag=GetSRKVal(ss, Mu, Ka, &sigma, &coeff,
		    &al1sM1sM2, &al1ssM2, &al1ssM1, &al2sM2sM3,
		    &al2sM1sM3, &al2ssM3, &al2ssM2, &al2ssM1);
  if (1==errflag) {
    printf("Error: ss is not among our selection numbers!\n");
    printf("       it must satisfy 4<=ss<=22, or it must be\n");
    printf("       24, 26, 28, 30, 32, 35, 38, 41, 45, 49,\n");
    printf("       53, 58, 63, 68, 74, 80, 87, 95 or 104.\n");
    exit(1);
  }
  sM2=ss-2;
  sM3=ss-3;

  /* See Appendix A in a technical report [Burrage_and_Komori:2010]. */
  al0sM1sM2=Mu[sM3];
  /* See Subsection 5.1 in [Burrage_Komori:2010]. */
  al2sM1sM2=al2sM2sM3;

  /* See Subsection 5.2 in [Burrage_Komori:2010]. */
  b2s=1.0;
  al6ssM2=1.0/(4.0*b2s);
  al6sM1sM2=-al6ssM2;
  b2sM1=-b2s;

  if (0==static_flag) {
    if (0!=traj%MaxCoreNum) {
      printf("Error: Number of trajects must be %d multiple!", MaxCoreNum);
      exit(1);
    }
    traj_mini=traj/MaxCoreNum;
    static_flag=1;
  }

/*#pragma omp parallel for*/
  for (ii_par=0; ii_par<MaxCoreNum; ii_par++) {
    unsigned long itr;
    int ii, jj, ll, ibase3p, ibase2p, ibase, ibase_work, ibase_work_step;
    double *yn, *yn1, *fn, *pjM2, *pjM1, *pj, *Y00sM2, *Y00sM1, *Y00s, 
      tmp1, tmp2, tmp3, tmp4, tmp5, tmp6;
    double **gn, *dummy, **YjjsM3, **YjlsM2, **YjjsM2, **YjjsM1,
      **YkjsM1, **Yjjs, **Ykjs;
    char *wj, *wtj;

    if (NULL == (gn = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (YjjsM3 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    /* This means Y^(j,1), ..., Y^(j,j-1), Y^(j,j+1), ..., Y^(j,wdim) */
    if (NULL == (YjlsM2 = (double **)malloc(sizeof(dummy)*wdim*wdimM1))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (YjjsM2 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (YjjsM1 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (YkjsM1 = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (Yjjs = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }
    if (NULL == (Ykjs = (double **)malloc(sizeof(dummy)*wdim))) {
      printf("malloc error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
      exit(0);
    }

    ibase_work_step=8+7*wdim+wdim*wdimM1;
    ibase_work=(ibase_work_step*ydim)*ii_par; /**/
    ii=0;
    yn=&work[ibase_work+ii];
    ii+=ydim;
    fn=&work[ibase_work+ii];
    ii+=ydim;
    pjM2=&work[ibase_work+ii];
    ii+=ydim;
    pjM1=&work[ibase_work+ii];
    ii+=ydim;
    pj=&work[ibase_work+ii];
    ii+=ydim;
    Y00sM2=&work[ibase_work+ii];
    ii+=ydim;
    Y00sM1=&work[ibase_work+ii];
    ii+=ydim;
    Y00s=&work[ibase_work+ii];
    ii+=ydim;
    for (jj=0; jj<wdim; jj++) {
      gn[jj]=&work[ibase_work+ii];
      ii+=ydim;
      YjjsM3[jj]=&work[ibase_work+ii];
      ii+=ydim;
      YjjsM2[jj]=&work[ibase_work+ii];
      ii+=ydim;
      YjjsM1[jj]=&work[ibase_work+ii];
      ii+=ydim;
      YkjsM1[jj]=&work[ibase_work+ii];
      ii+=ydim;
      Yjjs[jj]=&work[ibase_work+ii];
      ii+=ydim;
      Ykjs[jj]=&work[ibase_work+ii];
      ii+=ydim;
    }
    for (jj=0; jj<wdim*wdimM1; jj++) {
      YjlsM2[jj]=&work[ibase_work+ii];
      ii+=ydim;
    }

    ibase=(ydim*traj_mini)*ii_par;
    ibase3p=(wdim*traj_mini)*ii_par;
    ibase2p=(wdimM1*traj_mini)*ii_par;
    for(itr=0; itr<traj_mini; itr++) {
      wj=&ran3p[ibase3p];
      ibase3p+=wdim;
      wtj=&ran2p[ibase2p];
      ibase2p+=wdimM1;
      yn1=&ynew[ibase];
      for(ii=0; ii<ydim; ii++) {
	yn[ii]=yvec[ibase+ii];
      }
      ibase+=ydim;
      /* See Section 2 in [Burrage_Komori:2010]. */
      /* Calculations for K_{s-2}-h*mu_{s-2}*f(K_{s-3})  */
      ffunc(yn,fn);
      tmp1=step*Mu[0];
      for(ii=0; ii<ydim; ii++) {
	pjM2[ii]=yn[ii];
	pjM1[ii]=yn[ii]+tmp1*fn[ii];
      }
      for(jj=1; jj<sM2; jj++) {
	ffunc(pjM1,fn);
	tmp1=step*Mu[jj];
	tmp2=1+Ka[jj-1];
	tmp3=-Ka[jj-1];
	for (ii=0; ii<ydim; ii++) {
	  pj[ii]=tmp1*fn[ii]+tmp2*pjM1[ii]+tmp3*pjM2[ii];
	  if (jj<sM3) {
	    pjM2[ii]=pjM1[ii];
	    pjM1[ii]=pj[ii];
	  }
	}
      }
      for (ii=0; ii<ydim; ii++) {
	pj[ii]-=tmp1*fn[ii];
      }

      for (jj=0; jj<wdim; jj++) {
	gfunc_gene(pjM1,jj,gn[jj]);
      }
      for (ll=0; ll<wdim; ll++) {
	switch (wj[ll]) {
	case 0:
	  for (ii=0; ii<ydim; ii++) {
	    YjjsM3[ll][ii]=0;
	    for (jj=ll; jj<wdimM1; jj++) {
	      YjlsM2[ll*wdimM1+jj][ii]=0;
	      YjlsM2[(jj+1)*wdimM1+ll][ii]=0;
	    }
	  }
	  break;
	case 1:
	  for (ii=0; ii<ydim; ii++) {
	    YjjsM3[ll][ii]=sq3step*gn[ll][ii];
	  }
	  for (jj=ll; jj<wdimM1; jj++) {
	    if (0<wtj[jj]) {
	      for (ii=0; ii<ydim; ii++) {
		/* This means Y^(l,j)_{s-2} for j>l. */
		/* See also a comment in the part where malloc was called
		   for YjlsM2. */
		YjlsM2[ll*wdimM1+jj][ii]=sq3step*gn[jj+1][ii];
		/* This means Y^(j,l)_{s-2} for j>l. */
		YjlsM2[(jj+1)*wdimM1+ll][ii]=-sq3step*gn[ll][ii];
	      }
	    } else {
	      for (ii=0; ii<ydim; ii++) {
		YjlsM2[ll*wdimM1+jj][ii]=-sq3step*gn[jj+1][ii];
		YjlsM2[(jj+1)*wdimM1+ll][ii]=sq3step*gn[ll][ii];
	      }
	    }
	  }
	  break;
	default:
	  for (ii=0; ii<ydim; ii++) {
	    YjjsM3[ll][ii]=-sq3step*gn[ll][ii];
	  }
	  for (jj=ll; jj<wdimM1; jj++) {
	    if (0<wtj[jj]) {
	      for (ii=0; ii<ydim; ii++) {
		YjlsM2[ll*wdimM1+jj][ii]=-sq3step*gn[jj+1][ii];
		YjlsM2[(jj+1)*wdimM1+ll][ii]=sq3step*gn[ll][ii];
	      }
	    } else {
	      for (ii=0; ii<ydim; ii++) {
		YjlsM2[ll*wdimM1+jj][ii]=sq3step*gn[jj+1][ii];
		YjlsM2[(jj+1)*wdimM1+ll][ii]=-sq3step*gn[ll][ii];
	      }
	    }
	  }
	}
      } /* End of the loop for ll*/
      
      for (ii=0; ii<ydim; ii++) {
	tmp1=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=YjjsM3[jj][ii];
	}
	Y00sM2[ii]=pjM1[ii]+al2sM2sM3*tmp1;
      }
      ffunc(Y00sM2,fn);
      for (ii=0; ii<ydim; ii++) {
	Y00sM2[ii]=step*fn[ii];
      }

      for (jj=0; jj<wdim; jj++) {
	for (ii=0; ii<ydim; ii++) {
	  YjjsM2[jj][ii]=pj[ii]+al3sM2sM3*YjjsM3[jj][ii];
	}
	gfunc_gene(YjjsM2[jj],jj,gn[jj]);
	switch (wj[jj]) {
	case 0:
	  for (ii=0; ii<ydim; ii++) {
	    YjjsM2[jj][ii]=0;
	  }
	  break;
	case 1:
	  for (ii=0; ii<ydim; ii++) {
	    YjjsM2[jj][ii]=sq3step*gn[jj][ii];
	  }
	  break;
	default:
	  for (ii=0; ii<ydim; ii++) {
	    YjjsM2[jj][ii]=-sq3step*gn[jj][ii];
	  }
	}
      }
      
      for (ii=0; ii<ydim; ii++) {
	tmp1=0;
	tmp2=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=YjjsM3[jj][ii];
	  tmp2+=YjjsM2[jj][ii];
	}
	Y00sM1[ii]=pj[ii]+al0sM1sM2*Y00sM2[ii]+al2sM1sM3*tmp1+al2sM1sM2*tmp2;
      }
      ffunc(Y00sM1,fn);
      for (ii=0; ii<ydim; ii++) {
	Y00sM1[ii]=step*fn[ii];
      }

      for (jj=0; jj<wdim; jj++) {
	for (ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  tmp2=0;
	  for (ll=0; ll<wdim; ll++) {
	    if (ll!=jj) {
	      tmp1+=YjjsM3[ll][ii];
	      tmp2+=YjjsM2[ll][ii];
	    }
	  }
	  YjjsM1[jj][ii]=pjM1[ii]+al1sM1sM2*Y00sM2[ii]
	    +al3sM1sM3*YjjsM3[jj][ii]+al3sM1sM2*YjjsM2[jj][ii]
	    +al4sM1sM3*tmp1+al4sM1sM2*tmp2;
	}
	gfunc_gene(YjjsM1[jj],jj,gn[jj]);
	switch (wj[jj]) {
	case 0:
	  for (ii=0; ii<ydim; ii++) {
	    YjjsM1[jj][ii]=0;
	  }
	  break;
	case 1:
	  for (ii=0; ii<ydim; ii++) {
	    YjjsM1[jj][ii]=sq3step*gn[jj][ii];
	  }
	  break;
	default:
	  for (ii=0; ii<ydim; ii++) {
	    YjjsM1[jj][ii]=-sq3step*gn[jj][ii];
	  }
	}
      }

      for (jj=0; jj<wdim; jj++) {
	for (ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  for (ll=0; ll<wdimM1; ll++) {
	    tmp1+=YjlsM2[jj*wdimM1+ll][ii];
	  }
	  YkjsM1[jj][ii]=pjM1[ii]+al6sM1sM2*tmp1;
	}
	gfunc_gene(YkjsM1[jj],jj,gn[jj]);
	for (ii=0; ii<ydim; ii++) {
	  YkjsM1[jj][ii]=sqstep*gn[jj][ii];
	}
      }
      
      for (ii=0; ii<ydim; ii++) {
	tmp1=0;
	tmp2=0;
	tmp3=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=YjjsM3[jj][ii];
	  tmp2+=YjjsM2[jj][ii];
	  tmp3+=YjjsM1[jj][ii];
	}
	Y00s[ii]=pj[ii]+al0sM1sM2*Y00sM2[ii]+sigma*Y00sM1[ii]
	  +al2ssM3*tmp1+al2ssM2*tmp2+al2ssM1*tmp3;
      }
      ffunc(Y00s,fn);
      for (ii=0; ii<ydim; ii++) {
	Y00s[ii]=step*fn[ii];
      }

      for (jj=0; jj<wdim; jj++) {
	for (ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  tmp2=0;
	  for (ll=0; ll<wdim; ll++) {
	    if(ll!=jj) {
	      tmp1+=YjjsM3[ll][ii];
	      tmp2+=YjjsM2[ll][ii];
	    }
	  }
	  Yjjs[jj][ii]=pjM1[ii]+al1ssM2*Y00sM2[ii]+al1ssM1*Y00sM1[ii]
	    +al3ssM3*YjjsM3[jj][ii]+al3ssM2*YjjsM2[jj][ii]
	    +al3ssM1*YjjsM1[jj][ii]
	    +al4ssM3*tmp1+al4ssM2*tmp2;
	}
	gfunc_gene(Yjjs[jj],jj,gn[jj]);
	switch (wj[jj]) {
	case 0:
	  for (ii=0; ii<ydim; ii++) {
	    Yjjs[jj][ii]=0;
	  }
	  break;
	case 1:
	  for (ii=0; ii<ydim; ii++) {
	    Yjjs[jj][ii]=sq3step*gn[jj][ii];
	  }
	  break;
	default:
	  for (ii=0; ii<ydim; ii++) {
	    Yjjs[jj][ii]=-sq3step*gn[jj][ii];
	  }
	}
      }

      for (jj=0; jj<wdim; jj++) {
	for (ii=0; ii<ydim; ii++) {
	  tmp1=0;
	  for (ll=0; ll<wdimM1; ll++) {
	    tmp1+=YjlsM2[jj*wdimM1+ll][ii];
	  }
	  Ykjs[jj][ii]=pjM1[ii]+al6ssM2*tmp1;
	}
	gfunc_gene(Ykjs[jj],jj,gn[jj]);
	for (ii=0; ii<ydim; ii++) {
	  Ykjs[jj][ii]=sqstep*gn[jj][ii];
	}
      }
      
      for (ii=0; ii<ydim; ii++) {
	tmp1=tmp2=tmp3=tmp4=tmp5=tmp6=0;
	for (jj=0; jj<wdim; jj++) {
	  tmp1+=YjjsM3[jj][ii];
	  tmp2+=YjjsM2[jj][ii];
	  tmp3+=YjjsM1[jj][ii];
	  tmp4+=Yjjs[jj][ii];
	  tmp5+=YkjsM1[jj][ii];
	  tmp6+=Ykjs[jj][ii];
	}
	yn1[ii]=pj[ii]+al0sM1sM2*Y00sM2[ii]+sigma*(Y00sM1[ii]+Y00s[ii])
	  +coeff*(Y00s[ii]-Y00sM1[ii])
	  +b1sM3*tmp1+b1sM2*tmp2+b1sM1*tmp3+b1s*tmp4+b2sM1*tmp5+b2s*tmp6;
      }
    } /* End of loop for itr */
    free(gn);
    free(YjjsM3);
    free(YjlsM2);
    free(YjjsM2);
    free(YjjsM1);
    free(YkjsM1);
    free(Yjjs);
    free(Ykjs);
  }
  return 0;
}
