/* Filename: main.c */
/* 17-Nov-2010 */
/********************************************/
/* This solves Stratonovich SDEs            */
/*    dy=[A-(B1^2+B2^2+B3^2)/2]*y*dt        */
/*        +B1*y*dw1+B2*y*dw2+B3*y*dw3       */
/* by SROCK2, where                         */
/*    A=[0 1; -3 -2], B1=[1/2 1; -1/2 1/4], */
/*    B2=[0 1/4; 1 0], B3=[0 1/8; 1/4 0].   */
/********************************************/
#include <stdio.h>
#include <math.h>
#include <direct.h>

/* Terminal time */
#define  T_end               1.0

/* A step size */
#define  STEPLENGCONST    1.0/8.0

#define  YDIM         2

/* Setting for SDE */
#define  B1_11       0.5
#define  B1_12       1.0
#define  B1_21      -0.5
#define  B1_22       0.25
#define  B2_11       0.0
#define  B2_12       0.25
#define  B2_21       1.0
#define  B2_22       0.0
#define  B3_11       0.0
#define  B3_12       0.125
#define  B3_21       0.25
#define  B3_22       0.0
#define  BETA       -3.0
#define  GANMA      -2.0
#define  ONE         1.0
double    ModiTerm11 = -(B1_11*B1_11+B1_12*B1_21
			+B2_11*B2_11+B2_12*B2_21
			+B3_11*B3_11+B3_12*B3_21)/2.0;
double    ModiTerm12 = -(B1_11*B1_12+B1_12*B1_22
			+B2_11*B2_12+B2_12*B2_22
			+B3_11*B3_12+B3_12*B3_22)/2.0;
double    ModiTerm21 = -(B1_21*B1_11+B1_22*B1_21
			+B2_21*B2_11+B2_22*B2_21
			+B3_21*B3_11+B3_22*B3_21)/2.0;
double    ModiTerm22 = -(B1_21*B1_12+B1_22*B1_22
			+B2_21*B2_12+B2_22*B2_22
			+B3_21*B3_12+B3_22*B3_22)/2.0;

#define  NWinMax    30000UL /* >=WDIM*TRAJECT */
#define  TRAJECT    10000UL /* 100000UL */
#define  NArrayMax  30000UL /* >=YDIM*TRAJECT */
/* Total number of trajectries is TRAJECT*BATCH_NUM */
#define  BATCH_NUM 1 /* 1, 160, 2560 */

#define  WDIM         3

/* A drift coefficient  */
static void ffunc(double ynvec[],double foutput[]);

/* Diffusion coefficients */
static void gfunc_gene(double ynvec[],int i_th,double goutput[]);

/* Prallelized SROCK2 */
extern int OMP_wo2_srock_for_NonCommSDEs_WinMulti(int ydim,
						  unsigned long traj,
						  double *yvec,
						  double step,
						  char *ran3p, char *ran2p,
						  void (*ffunc)(),
						  void (*gfunc_gene)(),
						  int wdim, int ss,
						  double work[], double *ynew);

void init_genrand(unsigned long s);
int ran_gene_using_genrand_int32(int traject, int wdim,
				 char ran2p[], char ran3p[]);

int main(void) {
  /* For large memory area, the following variables are allocated
     in static area. */
  static double yvec[NArrayMax], ynew[NArrayMax];
  static char ran2p[NWinMax], ran3p[NWinMax];
  
  unsigned long seed = 5489UL; /* 5489L, 0UL */
  int traject=TRAJECT,ydim=YDIM, wdim=WDIM, stagenum, i, j, ibase, i_batch;
  double stepleng,
    expect[YDIM], tpoint, eps, yinit[YDIM], batch_expect[BATCH_NUM][YDIM],
    disper[YDIM], batch_disper[BATCH_NUM][YDIM], tmp, tmp1,
    work[8*(8+WDIM*(6+WDIM))*YDIM];

  stagenum=4; /* 4, 12, 16, 20, 41, 104 and so on. */
  
  yinit[0] = 1.0;
  yinit[1] = 0.0;
  
  if (NArrayMax < YDIM*TRAJECT) {
    printf("Error: NArrayMax is too small!\n");
    exit(0);
  }
  if (NWinMax < WDIM*TRAJECT) {
    printf("Error: NWinMax is too small!\n");
    exit(0);
  }

  /* The following sets a seed. */
  init_genrand(seed);

  stepleng=STEPLENGCONST;
  eps = 0.1*stepleng;
  /* Begining of a loop for i_batch */
  for(i_batch=0;i_batch<BATCH_NUM;i_batch++) {
    for(i=ibase=0;i<traject;i++) {/* Initialization for yvec */
      for(j=0;j<ydim;j++) {
	yvec[ibase+j] = yinit[j];
      }
      ibase+=ydim;
    }
    
    tpoint = 0.0; /* Initialization for tpoint */
    
    /* tpoint keeps changing until it reaches T_end. */
    while(eps<fabs(T_end-tpoint)) {
      /* A loop for i to generate random numbers for one step. */
      if(0!=ran_gene_using_genrand_int32(traject, wdim, ran2p, ran3p)) {
	printf("Error in ran_gene_using_genrand_init32!");
	exit(1);
      }
      
      /* Calculations for ynew by SROCK2 */
      if(0 != OMP_wo2_srock_for_NonCommSDEs_WinMulti(ydim, traject, yvec,
						     stepleng,
						     ran3p, ran2p,
						     ffunc, gfunc_gene,
						     wdim, stagenum,
						     work, ynew)) {
	printf("error in OMP_wo2_srock_for_NonCommSDEs_WinMulti\n");
	return 1;
      }
      
      for(i=ibase=0;i<traject;i++) {/* Update on yvec */
	for(j=0;j<ydim;j++) {
	  yvec[ibase+j] = ynew[ibase+j];
	}
	ibase+=ydim;
      }
      tpoint += stepleng;           /* Update on tpoint */
    }
    
    /* Calculations fo expectation in i_batch */
    for(i=0;i<ydim;i++)
      expect[i] = 0.0;
    for(i=ibase=0;i<traject;i++) {
      for(j=0;j<ydim;j++) {
	expect[j] += yvec[ibase+j];
      }
      ibase+=ydim;
    }
    for(i=0;i<ydim;i++)
      expect[i]/=traject;
    
    /* Calculations for dispersion in i_batch */
    for(i=0;i<ydim;i++)
      disper[i] = 0.0; 
    for(i=ibase=0;i<traject;i++) {
      for(j=0;j<ydim;j++) {
	disper[j] += (yvec[ibase+j] - expect[j])*(yvec[ibase+j] - expect[j]);
      }
      ibase+=ydim;
    }
    for(i=0;i<ydim;i++)
      disper[i]/=(traject-1);
    
    for(i=0;i<ydim;i++)
      batch_expect[i_batch][i]=expect[i];
    
    for(i=0;i<ydim;i++) {
      batch_disper[i_batch][i]=disper[i];
    }
  } /* End of the loop for i_batch */
  
  /* Calculations for expectation */
  for(i=0;i<ydim;i++) {
    expect[i] = 0.0;
    for(i_batch=0;i_batch<BATCH_NUM;i_batch++) {
      expect[i] += batch_expect[i_batch][i];
    }
    expect[i]/=BATCH_NUM;
  }
  
  /* Calculations for dispersion */
  for(i=0;i<ydim;i++) {
    tmp=tmp1=0.0;
    for(i_batch=0;i_batch<BATCH_NUM;i_batch++) {
      tmp+=batch_disper[i_batch][i];
      tmp1+=(batch_expect[i_batch][i] - expect[i])
	*(batch_expect[i_batch][i] - expect[i]);
    }
    disper[i]=((traject-1)*tmp+traject*tmp1)/(traject*BATCH_NUM-1);
  }
  
  for(i=0;i<ydim;i++)	{
    printf("expect%d: %17.14lf\n",i,expect[i]);
  }
  for(i=0;i<ydim;i++) {
    printf("disper%d: %17.14lf\n",i,disper[i]);
  }
  return 0;
}

static void ffunc(double ynvec[],double foutput[])
{
  foutput[0] = (ModiTerm11)*ynvec[0] + (ONE+(ModiTerm12))*ynvec[1];
  foutput[1] = (BETA+(ModiTerm21))*ynvec[0] + (GANMA + (ModiTerm22))*ynvec[1];
}

static void gfunc_gene(double ynvec[],int i_th,double goutput[])
{
  switch(i_th) {
  case 0:
    goutput[0] = B1_11*ynvec[0] + B1_12*ynvec[1];
    goutput[1] = B1_21*ynvec[0] + B1_22*ynvec[1];
    break;
  case 1:
    goutput[0] = B2_11*ynvec[0] + B2_12*ynvec[1];
    goutput[1] = B2_21*ynvec[0] + B2_22*ynvec[1];
    break;
  case 2:
    goutput[0] = B3_11*ynvec[0] + B3_12*ynvec[1];
    goutput[1] = B3_21*ynvec[0] + B3_22*ynvec[1];
    break;
  default:
    printf("Error_in_gfunc_gene: invalid diffusion coefficient was selected!\n");
    exit(0);
  }
}
